"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
class Asynctor {
    constructor(loader, tag = '') {
        this.__buffer = [];
        this.__state = 'init';
        this.__tag = tag;
        if (/this\.(?!_)\S+\(/.test(loader.toString())) {
            throw new Error(`Do not call public member methods in loader. It will cause cycle lock.`);
        }
        loader = loader.bind(this);
        process.nextTick(() => {
            this.__load(loader);
        });
        return new Proxy(this, {
            get(target, key, receiver) {
                const prop = Reflect.get(target, key, receiver);
                if (typeof prop !== 'function') {
                    return prop;
                }
                const fn = prop.bind(target);
                if (fn.constructor.name !== 'AsyncFunction') {
                    throw new Error(`The method [${key}] should be an async function`);
                }
                if (target.__state === 'succeed' || prop.name[0] === '_') {
                    return fn;
                }
                else {
                    target.__load(loader);
                    return async (...args) => {
                        return await new Promise((resolve, reject) => {
                            target.__buffer.push({
                                resolve,
                                reject,
                                fn,
                                args,
                            });
                        });
                    };
                }
            },
        });
    }
    __load(loader) {
        if (!['failed', 'init'].includes(this.__state)) {
            return;
        }
        this.__state = 'loading';
        if (!loader || loader.constructor.name !== 'AsyncFunction') {
            throw new Error('[Lazy Loader] loader must be an AsyncFunction');
        }
        loader().then(() => {
            this.__state = 'succeed';
            this.__onLoadSucceeded();
        }).catch(error => {
            this.__state = 'failed';
            this.__onLoadFailed(error);
        });
    }
    __onLoadSucceeded() {
        while (this.__buffer.length > 0) {
            const { resolve, reject, fn, args } = this.__buffer.shift();
            if (fn.constructor.name === 'AsyncFunction') {
                fn(...args).then(resolve).catch(reject);
            }
            else {
                try {
                    resolve(fn(...args));
                }
                catch (err) {
                    reject(err);
                }
            }
        }
    }
    __onLoadFailed(error) {
        this.__buffer.forEach(({ reject }) => {
            reject(new Error(`[Asynctor][${this.constructor.name}:${this.__tag}] loaded fail. Error ${error.message}`));
        });
    }
}
exports.default = Asynctor;
//# sourceMappingURL=index.js.map